/*
  Copyright (c) 2013, Matthias Schiffer <mschiffer@universe-factory.net>
  All rights reserved.

  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions are met:

    1. Redistributions of source code must retain the above copyright notice,
       this list of conditions and the following disclaimer.
    2. Redistributions in binary form must reproduce the above copyright notice,
       this list of conditions and the following disclaimer in the documentation
       and/or other materials provided with the distribution.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/


#include "context.hpp"
#include "config.hpp"
#include "network.hpp"
#include "node.hpp"

#include <algorithm>
#include <cassert>
#include <cstdio>


namespace MMSS {

static inline int snprintf_safe(char *buffer, size_t size, const char *format, ...) {
	std::va_list ap;
	va_start(ap, format);
	int ret = std::vsnprintf(buffer, size, format, ap);
	va_end(ap);

	return ret < 0 ? 0 : ret > size ? size : ret;
}

static inline const char* get_log_prefix(int log_level) {
	switch(log_level) {
	case LOG_CRIT:
		return "Fatal: ";
	case LOG_ERR:
		return "Error: ";
	case LOG_WARNING:
		return "Warning: ";
	case LOG_NOTICE:
		return "Info: ";
	case LOG_INFO:
		return "Verbose: ";
	case LOG_DEBUG:
		return "DEBUG: ";
	default:
		return "";
	}
}


void context_t::vlogf_orig(const node_t *orig, int priority, const char *format, std::va_list ap) {
	char buf[1024];
	size_t pos;

	if (orig)
		pos = snprintf_safe(buf, sizeof(buf), "[% 6u.%03u] %s: %s", now()/1000, now()%1000, orig->get_name().c_str(), get_log_prefix(priority));
	else
		pos = snprintf_safe(buf, sizeof(buf), "[% 6u.%03u] %s", now()/1000, now()%1000, get_log_prefix(priority));

	std::vsnprintf(buf+pos, sizeof(buf)-pos, format, ap);

	std::fprintf(stderr, "%s\n", buf);
}

void context_t::logf(int priority, const char *format, ...) {
	std::va_list ap;
	va_start(ap, format);
	vlogf_orig(nullptr, priority, format, ap);
	va_end(ap);
}

void context_t::run(int argc, char *argv[]) {
	if (argc != 2) {
		std::fprintf(stderr, "usage: %s protocol_module\n", argv[0]);
		std::exit(1);
	}

	std::shared_ptr<const protocol_t> proto = protocol_t::load(this, argv[1]);
	if (!proto)
		std::exit(1);

	std::shared_ptr<const config_t> conf = config_t::read(this, "babel_test.mmss");

	std::shared_ptr<network_t> net0 = std::make_shared<network_t>("net0");
	std::shared_ptr<network_t> net1 = std::make_shared<network_t>("net1");

	std::shared_ptr<node_t> node1 = node_t::create(this, "node1", 1, proto);
	std::shared_ptr<node_t> node2 = node_t::create(this, "node2", 2, proto);
	std::shared_ptr<node_t> node3 = node_t::create(this, "node3", 3, proto);

	std::list<std::shared_ptr<node_t>> nodes;

	nodes.push_back(node1);
	nodes.push_back(node2);
	nodes.push_back(node3);

	gmrf_addr_t addr1 = {{1}}, addr2 = {{2}}, addr3 = {{3}}, addr4 = {{4}};
	iface_t::add(node1, net0, "mmss0", &addr1);
	iface_t::add(node2, net0, "mmss0", &addr2);
	iface_t::add(node2, net1, "mmss1", &addr3);
	iface_t::add(node3, net1, "mmss1", &addr4);

	while (true) {
		if (now() > 10000000)
			break;

		int timeout = event_queue.timeout();

		if (timeout < 0) {
			fprintf(stderr, "nothing queued, deadlock occured.\n");
			break;
		}

		if (timeout > 0) {
			assert(!event_queue.get());

			advance(timeout);
			timeout = event_queue.timeout();
		}

		assert(timeout == 0);

		event_queue.get()->handle(this);
	}
}

}
